#include <stdlib.h>
#include <assert.h>

#include <chipmunk.h>

#if (CP_VERSION_MAJOR < 6)
#error "You need chipmunk 6"
#endif

#if !defined(cpAreaForBox)
#define cpAreaForBox(w,h) ((w) * (h))
#endif

#include "dynamic.h"

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
struct dynamic_s
{
    world_t * world;
    cpSpace * space;
};

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
static void shape_free(cpBody *body, cpShape *shape, cpSpace * space)
{
    assert(body && shape && space);

    /* remove shape from space (it automatically remove shape from body too) */
    if(cpSpaceContainsShape(space, shape))
    {
	cpSpaceRemoveShape(space, shape);
    }

    /* free shape */
    cpShapeFree(shape);
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
static void body_free(cpBody *body, cpSpace * space)
{
    assert(body && body->data && space);

    /* free each shape attached to this body */
    cpBodyEachShape(body, (cpBodyShapeIteratorFunc)shape_free, space);

    /* remove body from space */
    if(cpSpaceContainsBody(space, body))
    {
	cpSpaceRemoveBody(space, body);
    }

    /* unlink body from its item */
    item_t * item = body->data;
    item->dynamic_data = NULL;

    /* free body */
    cpBodyFree(body);

}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
int collision_begin_handler(cpArbiter *arb, struct cpSpace *space, void *data)
{
    CP_ARBITER_GET_SHAPES(arb, a, b);
    item_t *item_a = a->body->data;
    item_t *item_b = b->body->data;
    assert(item_a->collectable && item_b->collectable);
    assert(item_a->dynamic ^ item_b->dynamic);

    if(!item_a->dynamic)
    {
	item_a->collectable = 0;
	a->collision_type = 0;
    }    
    else if(!item_b->dynamic)
    {
	item_b->collectable = 0;
	b->collision_type = 0;
    }    
    return 1;
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
dynamic_t * dynamic_alloc(void)
{
    dynamic_t * dynamic;
    dynamic = calloc(1, sizeof(*dynamic));
    if(dynamic)
    {
	/* init chipmunk */
	cpInitChipmunk();
	/* instanciate the chipmunk's space */
	dynamic->space = cpSpaceNew();
	if(!dynamic->space)
	{
	    free(dynamic);
	    dynamic = NULL;
	}
    }
    return dynamic;
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
void dynamic_free(dynamic_t * dynamic)
{
    assert(dynamic);

    /* free chipmunk's space */
    cpSpaceFree(dynamic->space);
    dynamic->space = 0;

    free(dynamic);
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
void dynamic_init(dynamic_t * dynamic, world_t * world)
{
    assert(dynamic && world);

    /* max size of an item */ 
    cpFloat max_dim = 0;
    /* number of items */
    int count = 0;
 
    /* save a ref to the world */
    dynamic->world = world;

    /* for each items */
    item_t * item = dynamic->world->items_head;
    while(item)
    {
        /* compute area an moment */
	cpFloat area;
	cpFloat moment;
	if(item->shape == Circle)
	{
	    area = cpAreaForCircle(0, item->radius);
	    moment = cpMomentForCircle(area, 0, item->radius, cpvzero);
	}
	else if(item->shape == Box)
	{
	    area = cpAreaForBox(item->width, item->height);
	    moment = cpMomentForBox(area, item->width, item->height);
	}
	else
	{
	    continue;
	}

	/* create a body */
	cpBody * body;
	if(item->dynamic)
	{
	    body = cpBodyNew(area, moment);
	    /* only add the body to the space if it isn't static */
	    cpSpaceAddBody(dynamic->space, body);
	}
	else
	{
	    body = cpBodyNewStatic();
	}

        /* set body's position and angle */
	cpBodySetPos(body, cpv(item->position[0], item->position[1]));
	cpBodySetAngle(body, item->angle);

	/* link the body with its item */
	cpBodySetUserData(body, item);
	item->dynamic_data = body;

	/* create a shape : the rule is only one shape per body */
	cpShape *shape;
	if(item->shape == Circle)
	{
	    shape = cpCircleShapeNew(body, item->radius, cpvzero);
	}
	else //if(item->shape == Box)
	{
	    shape = cpBoxShapeNew(body, item->width, item->height);
	}
	cpSpaceAddShape(dynamic->space, shape);

	/* setup shape properties */
	cpShapeSetElasticity(shape, 0.5); 
	cpShapeSetFriction(shape, 0.4);
	cpShapeSetCollisionType(shape, item->collectable?1:0);

        /* coherency checks */
	assert(item->dynamic ^ cpBodyIsStatic(body));
	assert(item->dynamic ^ cpSpaceContainsBody(dynamic->space, body));

	/* check max size */
	if(max_dim < item->width)
	{
	    max_dim = item->width;
	}
	if(max_dim < item->height)
	{
	    max_dim = item->height;
	}
	count++;

	/* next one ! */
	item = item->next;
    }

    /* setup space properties */
    cpSpaceReindexStatic(dynamic->space);

    cpSpaceUseSpatialHash(dynamic->space, max_dim * 2.0, count * 10);

    cpSpaceAddCollisionHandler(dynamic->space, 1, 1, collision_begin_handler, 0, 0, 0, 0);

    cpSpaceSetGravity(dynamic->space, cpv(dynamic->world->gravity[0], dynamic->world->gravity[1]));

    assert((dynamic->world->gravity[0] == 0) && (dynamic->world->gravity[1] < 0));
}


/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
void dynamic_kill(dynamic_t * dynamic)
{
    assert(dynamic);

    /* for each items */
    item_t * item = dynamic->world->items_head;
    while(item)
    {
	cpBody *body = item->dynamic_data;
	assert(body);

	body_free(body, dynamic->space);
	
	item = item->next;
    }
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
void dynamic_update(dynamic_t * dynamic, double dt)
{
    assert(dynamic);

    /* one dynamic step */
    cpSpaceStep(dynamic->space, dt);

    /* for each items */
    item_t * item = dynamic->world->items_head;
    while(item)
    {
	/* memorize the next item in case we free this one */
	item_t * next = item->next;

	/* check the body */
	cpBody *body = item->dynamic_data;
	assert(body);
	
	/* if it is a moving body */
	if(!cpBodyIsStatic(body))
	{
	    /* get body's position and angle */
	    cpVect pos = cpBodyGetPos(body);
	    cpFloat a = cpBodyGetAngle(body);
	    
	    /* update item position and angle */
	    item->position[0] = pos.x;
	    item->position[1] = pos.y;
	    item->angle = a;

	    /* check if it isn't outside */
	    /* FIXME : only works if the gravity is negative and only on "y" */
	    if(((item->position[1] + item->width)  < 0) && 
	       ((item->position[1] + item->height) < 0))
	    {
		/* free the body */
		body_free(body, dynamic->space);
		
		/* IMPORTANT : never let an item without a body !!! */
		item_free(item, &dynamic->world->items_head);
	    }
	}

	/* next one ! */
	item = next;
    }
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
void dynamic_remove(dynamic_t * dynamic, double x, double y)
{
    assert(dynamic);

    /* is there any shape that hit */
    cpShape * shape = cpSpacePointQueryFirst(dynamic->space, cpv(x, y), CP_ALL_LAYERS, CP_NO_GROUP);
    if(shape)
    {
	cpBody * body = shape->body;
	assert(body);

	item_t * item = body->data;
	assert(item);

	/* if we can remove this item */
	if(item->removable)
	{
	    /* free the body */
	    body_free(body, dynamic->space);
		
	    /* IMPORTANT : never let an item without a body !!! */
	    item_free(item, &dynamic->world->items_head);
	}
    }
}
