/**
 *
 * Ressource :
 *  - SDL        : http://sdl.beuc.net/sdl.wiki/SDL_API
 *    SDL_image  : http://jcatki.no-ip.org:8080/SDL_image
 *    SDL_gfx    : http://www.ferzkopp.net/joomla/content/view/19/14/
 *  
 *  - chipmunk   : http://files.slembcke.net/chipmunk/release/ChipmunkLatest-Docs
 *
 *  - opengl     : http://glprogramming.com/red/
 *
 *  - expat      : http://equinox.homelinux.org/viewcvs/ediconv/expat/reference.html?view=co
 *
 */
#include <stdio.h>
#include <stdlib.h>

#include <SDL/SDL.h>

#include "item.h"
#include "world.h"
#include "render.h"
#include "dynamic.h"

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
#define DYNAMIC_UPDATE 0
static const Uint32 dynamic_timer_interval_ms = 16;
static const SDL_Event dynamic_event = {.user.type=SDL_USEREVENT, .user.code=DYNAMIC_UPDATE};
static SDL_TimerID dynamic_timer = 0;

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
#define RENDER_UPDATE 1
static const Uint32 render_timer_interval_ms = 42;
static const SDL_Event render_event = {.user.type=SDL_USEREVENT, .user.code=RENDER_UPDATE};
static SDL_TimerID render_timer = 0;

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
static Uint32 push_event_callback(Uint32 interval, void *param)
{
    SDL_Event *event = (SDL_Event *)param;
    SDL_PushEvent(event);
    return interval;
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
static int sdl_init(int width, int height)
{
    const SDL_VideoInfo* info = NULL;
    int bpp = 0;
    int flags = 0;

    /* init sdl subsystems */
    if( SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0 ) 
    {
	fprintf( stderr, "SDL video and timer initialization failed: %s\n", SDL_GetError( ) );
	return -1;
    }
    
    /* init opengl options */
    SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
    SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
    SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
    SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

    /* get bpp of the screen */
    info = SDL_GetVideoInfo( );
    if( !info ) 
    {
        fprintf( stderr, "Video query failed: %s\n", SDL_GetError( ) );
        return -1;
    }
    bpp = info->vfmt->BitsPerPixel;

    flags = SDL_OPENGL | SDL_HWSURFACE | SDL_DOUBLEBUF;

    /* setup video */
    if( SDL_SetVideoMode(width, height, bpp, flags ) == 0 ) 
    {
        fprintf( stderr, "Video mode set failed: %s\n", SDL_GetError( ) );
        return -1;
    }

    /* add a timer for triggering the dynamic */
    dynamic_timer = SDL_AddTimer(dynamic_timer_interval_ms, push_event_callback, (void *)&dynamic_event);

    /* add a timer for triggering the rendering */
    render_timer = SDL_AddTimer(render_timer_interval_ms, push_event_callback, (void *)&render_event);

    return 0;
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
static void sdl_kill(void)
{
    SDL_RemoveTimer(dynamic_timer);
    dynamic_timer = 0;

    SDL_RemoveTimer(render_timer);
    render_timer = 0;

    SDL_Quit();
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
int main(int argc, char **argv)
{
    world_t * world = NULL;
    render_t * render = NULL;
    dynamic_t * dynamic = NULL;

    // init world and items
    if(argc >= 2)
    {
	world = world_alloc_from_xml(argv[1]);
    }
    if(!world)
    {
	fprintf(stderr, "Will generate a random world\n");
	world = world_alloc_from_random();
	if(!world)
	{
	    return 1;
	}
    }

    // alloc subsystems
    render = render_alloc();
    dynamic = dynamic_alloc();

    // init subsystems
    sdl_init(world->width, world->height);
    render_init(render, world);
    dynamic_init(dynamic,  world);

#if !defined(NDEBUG)
    // timings check
    Uint32 dynamic_start_date;
    Uint32 dynamic_end_date;
    Uint32 dynamic_elapsed_time_ms;
    Uint32 render_start_date;
    Uint32 render_end_date;
    Uint32 render_elapsed_time_ms;

    // debug world
    world_print(world, stdout);
#endif

    // main loop
    int status = 0;
    int pause = 0;
    SDL_Event event; 
    while(SDL_WaitEvent(&event)) 
    {
        if(event.type == SDL_QUIT)
	{
	    break;
	}
	else
	{
	    switch(event.type) 
	    {
	    case SDL_MOUSEBUTTONUP:
		if(!pause && (event.button.button == SDL_BUTTON_LEFT))
		{
		    dynamic_remove(dynamic, event.button.x, world->height - event.button.y);
		}
		break;
	    case SDL_KEYUP:
		if(event.key.keysym.sym == SDLK_SPACE)
		{
		    pause = 1 - pause;
		}
		break;
	    case SDL_USEREVENT:
		if(!pause && (event.user.code == DYNAMIC_UPDATE))
		{
#if !defined(NDEBUG)
		    dynamic_start_date = SDL_GetTicks();
#endif
		    dynamic_update(dynamic, dynamic_timer_interval_ms / 1000.0);
#if !defined(NDEBUG)
		    dynamic_end_date = SDL_GetTicks();
		    dynamic_elapsed_time_ms = dynamic_end_date - dynamic_start_date;
		    if(dynamic_elapsed_time_ms > dynamic_timer_interval_ms)
		    {
			fprintf(stderr, "dynamic overrun (took %u ms)\n", dynamic_elapsed_time_ms); 
		    }
#endif
		}
		else if(event.user.code == RENDER_UPDATE)
		{
#if !defined(NDEBUG)
		    render_start_date = SDL_GetTicks();
#endif
		    render_update(render);
		    SDL_GL_SwapBuffers();
#if !defined(NDEBUG)
		    render_end_date = SDL_GetTicks();
		    render_elapsed_time_ms = render_end_date - render_start_date;
		    if(render_elapsed_time_ms > render_timer_interval_ms)
		    {
			fprintf(stderr, "render overrun (took %u ms)\n", render_elapsed_time_ms); 
		    }
#endif
		}
		break;
	    }
	    if((event.type == SDL_MOUSEBUTTONUP) || (event.type == SDL_USEREVENT))
	    {
		if(status <= 0)
		{
		    status = world_check(world);
		    if(0 < status)
		    {
			fprintf(stdout, "Win !\n");
		    }
		}
	    }
	}
    }

    // kill subsystems
    dynamic_kill(dynamic);
    render_kill(render);
    sdl_kill();

    // free subsystems
    dynamic_free(dynamic);
    render_free(render);

    // free world and items
    world_free(world);

    fprintf(stdout, "%s !\n", (status<=0)?"Game over":"Win");

    return 0;
}
