#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <expat.h>

#include "world.h"

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/

struct xml_handler_arg_s
{
    world_t * world;
    int err;
};
typedef struct xml_handler_arg_s xml_handler_arg_t;

static void XMLCALL xml_start_element_handler(void *user_data, const XML_Char *elt_name, const XML_Char **elt_attr)
{
    XML_Parser xml_parser = (XML_Parser)user_data;
    xml_handler_arg_t *args = (xml_handler_arg_t *)XML_GetUserData(xml_parser);

    world_t * world = args->world;
    assert(world);

    int i = 0;

    if(strcmp(elt_name, "world") == 0)
    {
	for(i = 0; elt_attr[i] != NULL; i += 2)
	{
	    char * endptr = NULL;
	    if(strcmp(elt_attr[i], "gravity_x") == 0)
	    {
		world->gravity[0] = strtod(elt_attr[i + 1], &endptr);
	    }
	    else if(strcmp(elt_attr[i], "gravity_y") == 0)
	    {
		world->gravity[1] = strtod(elt_attr[i + 1], &endptr);
	    }
	    else if(strcmp(elt_attr[i], "image") == 0)
	    {
		strncpy(world->image, elt_attr[i + 1], sizeof(world->image));
	    }
	    else if(strcmp(elt_attr[i], "color") == 0)
	    {
		unsigned color = strtoul(elt_attr[i + 1], &endptr, 16);
		world->color[0] = ((color >> 24) & 0xff) / (double)0xff;
		world->color[1] = ((color >> 16) & 0xff) / (double)0xff;
		world->color[2] = ((color >>  8) & 0xff) / (double)0xff;
		world->color[3] = ((color >>  0) & 0xff) / (double)0xff;
	    }
	    else if(strcmp(elt_attr[i], "width") == 0)
	    {
		world->width = strtod(elt_attr[i + 1], &endptr);
	    }
	    else if(strcmp(elt_attr[i], "height") == 0)
	    {
		world->height = strtod(elt_attr[i + 1], &endptr);
	    }
	    /* if the conversion failed */
	    if (endptr == elt_attr[i + 1]) 
	    {
		fprintf(stderr, "ato* error\n");
		break;
	    }
	}	    
    }
    else if(strcmp(elt_name, "item") == 0)
    {
	item_t * item = item_alloc(&world->items_head);
	if(item)
	{
	    for(i = 0; elt_attr[i] != NULL; i += 2)
	    {
		char * endptr = NULL;
		if(strcmp(elt_attr[i], "id") == 0)
		{
		    item->id = strtoul(elt_attr[i + 1], &endptr, 0);
		}
		else if(strcmp(elt_attr[i], "shape") == 0)
		{
		    if(strcmp(elt_attr[i + 1], "circle") == 0)
		    {
			item->shape = Circle;
		    }
		    else if(strcmp(elt_attr[i + 1], "box") == 0)
		    {
			item->shape = Box;
		    }
		}
		else if(strcmp(elt_attr[i], "x") == 0)
		{
		    item->position[0] = strtod(elt_attr[i + 1], &endptr);
		}
		else if(strcmp(elt_attr[i], "y") == 0)
		{
		    item->position[1] = strtod(elt_attr[i + 1], &endptr);
		}
		else if(strcmp(elt_attr[i], "angle") == 0)
		{
		    item->angle = strtod(elt_attr[i + 1], &endptr);
		}
		else if(strcmp(elt_attr[i], "radius") == 0)
		{
		    item->radius = strtod(elt_attr[i + 1], &endptr);
		    item->radius_copy = item->radius;
		}
		else if(strcmp(elt_attr[i], "width") == 0)
		{
		    item->width = strtod(elt_attr[i + 1], &endptr);
		}
		else if(strcmp(elt_attr[i], "height") == 0)
		{
		    item->height = strtod(elt_attr[i + 1], &endptr);
		}
		else if(strcmp(elt_attr[i], "color") == 0)
		{
		    unsigned color = strtoul(elt_attr[i + 1], &endptr, 16);
		    item->color[0] = ((color >> 24) & 0xff) / (double)0xff;
		    item->color[1] = ((color >> 16) & 0xff) / (double)0xff;
		    item->color[2] = ((color >>  8) & 0xff) / (double)0xff;
		    item->color[3] = ((color >>  0) & 0xff) / (double)0xff;
		}
		else if(strcmp(elt_attr[i], "dynamic") == 0)
		{
		    if((elt_attr[i + 1][0] != 'n') && (elt_attr[i + 1][0] != '0'))
		    {
			item->dynamic = 1;
		    }
		}
		else if(strcmp(elt_attr[i], "collectable") == 0)
		{
		    if((elt_attr[i + 1][0] != 'n') && (elt_attr[i + 1][0] != '0'))
		    {
			item->collectable = 1;
		    }
		}
		else if(strcmp(elt_attr[i], "removable") == 0)
		{
		    if((elt_attr[i + 1][0] != 'n') && (elt_attr[i + 1][0] != '0'))
		    {
			item->removable = 1;
		    }
		}
		/* if the conversion failed */
		if (endptr == elt_attr[i + 1]) 
		{
		    fprintf(stderr, "ato* error\n");
		    /* dont' forget to remove the item from the list ! */
		    item_free(item, &world->items_head);
		    break;
		}
	    }
	}
	else
	{
	    fprintf(stderr, "item allocation failed\n");
	}
    }
    if(elt_attr[i] != NULL)
    {
	fprintf(stderr, "can't read all '%s' attributes.\n", elt_name);
	args->err = 1;
    }
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
world_t * world_alloc_from_xml(const char * filename)
{
    assert(filename);

    /* open the file */
    FILE * fp = fopen(filename, "r");
    if(!fp)
    {
	fprintf(stderr, "fopen(\"%s\") failed : %s\n", filename, strerror(errno));
	return NULL;
    }

    world_t * world = calloc(1, sizeof(*world));
    if(world)
    {
	/* structure for exchanging information with the handler */
	xml_handler_arg_t xml_handler_arg;
	xml_handler_arg.world = world;
	xml_handler_arg.err = 0;

	/* expat xml parser */
	XML_Parser xml_parser = XML_ParserCreate(NULL);
	XML_SetElementHandler(xml_parser, xml_start_element_handler, NULL);
	XML_UseParserAsHandlerArg(xml_parser);
	XML_SetUserData(xml_parser, &xml_handler_arg);

	/* go for parsing */
	do
	{
	    char buffer[1024];
	    char *line = fgets(buffer, sizeof(buffer), fp);
	    if(line)
	    {
		if(!XML_Parse(xml_parser, line, strlen(line), feof(fp)))
		{
		    fprintf(stderr, "XML_Parse failed : %s\n", XML_ErrorString(XML_GetErrorCode(xml_parser)));
		    world_free(world);
		    world = NULL;
		    break;
		}
		else if(xml_handler_arg.err)
		{
		    fprintf(stderr, "xml_start_element_handler failed\n");
		    world_free(world);
		    world = NULL;
		    break;
		}
	    }
	} while(!feof(fp));

	XML_ParserFree(xml_parser);

	fclose(fp);
    }

    /* sanity check */
    if(world && world_check(world))
    {
	world_free(world);
	world = NULL;
    }

    return world;
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
world_t * world_alloc_from_random(void)
{
    static int count = 128;

    srand(0);

    world_t * world = calloc(1, sizeof(*world));
    if(world)
    {
	world->gravity[0] = 0;
	world->gravity[1] = -80;
	
	world->width = 800;
	world->height = 600;
	world->color[0] = 1.0;
	world->color[1] = 1.0;
	world->color[2] = 1.0;
	world->color[3] = 1.0;
 
	// generate a random list of items
	int i;
	for(i = 0; i < count; i++)
	{
	    item_t * item = item_alloc(&world->items_head);
	    if(item)
	    {
		item_shape_t shape;
		unsigned is_dynamic;
		unsigned is_removable;
		double color[4];
		double x;
		double y;
		double a;
		double w;
		double h;
    
		shape = rand() % 2;
		is_dynamic = rand() % 2;
		is_removable = rand() % 2;
		color[0] = rand() / (double)RAND_MAX;
		color[1] = rand() / (double)RAND_MAX;
		color[2] = rand() / (double)RAND_MAX;
		color[3] = 1.0;
		x = 20 + rand() % (640 - 20);
		y = 20 + rand() % (480 - 20);
		a = 0;
		w = 5 + (rand() % 15);
		h = 5 + (rand() % 15);

		item->id = i;
		item->dynamic = rand() % 2;
		item->collectable = rand() % 2;
		item->removable = item->collectable?(rand() % 2):0;
		item->position[0] = 20 + rand() % (unsigned)(world->width - 20);
		item->position[1] = 20 + rand() % (unsigned)(world->height - 20);
		item->angle = 0;

		item->shape = rand() % 2;
		if(item->shape == Circle)
		{
		    item->radius = 5 + (rand() % 15);
		    item->radius_copy = item->radius;
		}
		else
		{
		    item->width = 5 + (rand() % 15);
		    item->height = 5 + (rand() % 15);
		}
		item->color[0] = color[0];
		item->color[1] = color[1];
		item->color[2] = color[2];
		item->color[3] = color[3];
	    }
	}
    }
    return world;
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
void world_free(world_t * world)
{
    assert(world);

    while(world->items_head)
    {
	item_free(world->items_head, &world->items_head);
    }

    free(world);
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
int world_check(world_t * world)
{
    assert(world);

    unsigned collectable_removable = 0;
    unsigned collectable_dynamic   = 0;
    unsigned collectable_static    = 0;

    item_t * item = world->items_head;
    while(item)
    {
	if(item->collectable)
	{
	    /* count removable collectable */
	    if(item->removable)
	    {
		collectable_removable++;
	    }
	    
            /* count static and dynamic collectable */    
	    if(item->dynamic)
	    {
		collectable_dynamic++;
	    }
	    else
	    {
		collectable_static++;
	    }
	}
	item = item->next;
    }

    /* a collectable can't be removable */
    if(collectable_removable)
    {
	return -2;
    }
    /* there must be one and only one dynamic collectable */
    else if(collectable_dynamic != 1)
    {
	return -1;
    }
    /* still need to collect some static */
    else if(collectable_static) 
    {
	return 0;
    }
    /* if there isn't any more static collectable : it's won ! */
    else
    {
	return 1;
    }
}

/*****************************************************************************
 *                                                                           *
 *****************************************************************************/
void world_print(world_t * world, FILE * fp)
{
    if(world)
    {
	fprintf(fp, "World : g=(%f,%f), w=%f, h=%f\n",
		world->gravity[0],
		world->gravity[1],
		world->width,
		world->height);

	item_print(world->items_head, fp, 1);
    }
}

